"""scenario.py
 
Module contains the DevOpsScenario class which defines a configurable trading scenario
which can be tested on a nullchain service and deployed to an existing network. The
scenario consists of the following agents.

    • 1x market manager
    • 1x market maker 
    • 2x auction traders
    • 3x random traders
    • 5x momentum traders
    • 3x sensitive traders
 
"""

import numpy as np
from datetime import datetime, timedelta
from typing import Optional, Union, Dict, Callable, Any

from vega_sim.scenario.scenario import Scenario
from vega_sim.scenario.constants import Network
from vega_sim.null_service import VegaServiceNull
from vega_sim.network_service import VegaServiceNetwork
from vega_sim.environment.environment import (
    MarketEnvironmentWithState,
    NetworkEnvironment,
    Agent,
)
from vega_sim.scenario.common.utils.price_process import (
    LivePrice,
    get_historic_price_series,
)
from vega_sim.scenario.common.agents import (
    ExponentialShapedMarketMaker,
    OpenAuctionPass,
    MarketOrderTrader,
    PriceSensitiveLimitOrderTrader,
)
from vega_sim.scenario.configurable_market.agents import ConfigurableMarketManager
from vega_sim.api.market import MarketConfig
from vega_sim.devops.wallet import (
    MARKET_CREATOR_AGENT,
    MARKET_SETTLER_AGENT,
    MARKET_MAKER_AGENT,
    AUCTION_TRADER_AGENTS,
    RANDOM_TRADER_AGENTS,
    SENSITIVE_TRADER_AGENTS,
)
from vega_sim.devops.classes import (
    MarketMakerArgs,
    MarketManagerArgs,
    AuctionTraderArgs,
    RandomTraderArgs,
    SensitiveTraderArgs,
    SimulationArgs,
)


class DevOpsScenario(Scenario):
    def __init__(
        self,
        binance_code: str,
        market_manager_args: MarketManagerArgs,
        market_maker_args: MarketMakerArgs,
        auction_trader_args: AuctionTraderArgs,
        random_trader_args: RandomTraderArgs,
        sensitive_trader_args: SensitiveTraderArgs,
        simulation_args: Optional[SimulationArgs] = None,
        state_extraction_fn: Optional[
            Callable[[VegaServiceNull, Dict[str, Agent]], Any]
        ] = None,
        step_length_seconds: float = 10,
        market_name: Optional[str] = None,
    ):
        super().__init__(state_extraction_fn=state_extraction_fn)

        self.binance_code = binance_code

        self.market_manager_args = market_manager_args
        self.market_maker_args = market_maker_args

        self.auction_trader_args = auction_trader_args
        self.random_trader_args = random_trader_args
        self.sensitive_trader_args = sensitive_trader_args
        self.simulation_args = simulation_args

        self.step_length_seconds = step_length_seconds
        self.market_name = market_name

    def _get_historic_price_process(
        self,
        random_state,
    ) -> list:
        start = datetime.strptime(self.simulation_args.start_date, "%Y-%m-%d %H:%M:%S")

        if self.simulation_args.randomise_history:
            start = start + timedelta(days=int(random_state.choice(range(30))))

        end = start + timedelta(
            seconds=(self.simulation_args.n_steps)
            * self.simulation_args.granularity.value
        )

        price_process = get_historic_price_series(
            product_id=self.simulation_args.coinbase_code,
            granularity=self.simulation_args.granularity,
            interpolation=(
                f"{self.step_length_seconds}S"
                if self.step_length_seconds < self.simulation_args.granularity.value
                else self.simulation_args.granularity.value
            ),
            start=str(start),
            end=str(end),
        )

        return price_process

    def configure_agents(
        self,
        vega: Union[VegaServiceNull, VegaServiceNetwork],
        tag: str,
        random_state: Optional[np.random.RandomState] = None,
        **kwargs,
    ) -> Dict[str, Agent]:
        random_state = (
            random_state if random_state is not None else np.random.RandomState()
        )

        if kwargs.get("network", Network.FAIRGROUND) == Network.NULLCHAIN:
            self.price_process = self._get_historic_price_process(
                random_state=random_state
            )
        else:
            self.price_process = LivePrice(product=self.binance_code)

        if kwargs.get("run_background", True):
            # Setup agent for proposing and settling the market
            market_manager = ConfigurableMarketManager(
                proposal_wallet_name=MARKET_CREATOR_AGENT.wallet_name,
                proposal_key_name=MARKET_CREATOR_AGENT.key_name,
                termination_wallet_name=MARKET_SETTLER_AGENT.wallet_name,
                termination_key_name=MARKET_SETTLER_AGENT.key_name,
                market_config=MarketConfig(),
                market_name=self.market_name
                if self.market_name is not None
                else self.market_manager_args.market_name,
                market_code=self.market_manager_args.market_code,
                asset_name=self.market_manager_args.asset_name,
                asset_dp=self.market_manager_args.adp,
                initial_mint=self.market_manager_args.initial_mint,
                settlement_price=self.price_process[-1],
                tag=None,
            )

            # Setup agent for proving a market for traders
            market_maker = ExponentialShapedMarketMaker(
                wallet_name=MARKET_MAKER_AGENT.wallet_name,
                key_name=MARKET_MAKER_AGENT.key_name,
                market_name=self.market_name
                if self.market_name is not None
                else self.market_manager_args.market_name,
                asset_name=self.market_manager_args.asset_name,
                initial_asset_mint=self.market_maker_args.initial_mint,
                commitment_amount=self.market_maker_args.commitment_amount,
                market_kappa=self.market_maker_args.market_kappa,
                kappa=self.market_maker_args.order_kappa,
                num_levels=self.market_maker_args.order_levels,
                tick_spacing=self.market_maker_args.order_spacing,
                inventory_lower_boundary=self.market_maker_args.inventory_lower_boundary,
                inventory_upper_boundary=self.market_maker_args.inventory_upper_boundary,
                fee_amount=self.market_maker_args.fee_amount,
                num_steps=60 * 60 * 24 * 365,
                price_process_generator=iter(self.price_process),
                orders_from_stream=False,
                state_update_freq=10,
                tag=None,
            )

            # Setup agents for passing opening auction
            auction_pass_agents = [
                OpenAuctionPass(
                    wallet_name=party.wallet_name,
                    key_name=party.key_name,
                    market_name=self.market_name
                    if self.market_name is not None
                    else self.market_manager_args.market_name,
                    asset_name=self.market_manager_args.asset_name,
                    initial_asset_mint=self.auction_trader_args.initial_mint,
                    initial_price=self.price_process[0],
                    side=["SIDE_BUY", "SIDE_SELL"][i],
                    tag=i,
                )
                for i, party in enumerate(AUCTION_TRADER_AGENTS)
            ]

            # Setup agents for placing random market orders
            random_market_order_traders = [
                MarketOrderTrader(
                    wallet_name=party.wallet_name,
                    key_name=party.key_name,
                    market_name=self.market_name
                    if self.market_name is not None
                    else self.market_manager_args.market_name,
                    asset_name=self.market_manager_args.asset_name,
                    initial_asset_mint=self.random_trader_args.initial_mint,
                    buy_intensity=self.random_trader_args.order_intensity[i],
                    sell_intensity=self.random_trader_args.order_intensity[i],
                    base_order_size=self.random_trader_args.order_volume[i],
                    step_bias=self.random_trader_args.step_bias[i],
                    tag=i,
                )
                for i, party in enumerate(RANDOM_TRADER_AGENTS)
            ]

            # Setup agents for placing price-sensitive limit orders
            sensitive_limit_order_traders = [
                PriceSensitiveLimitOrderTrader(
                    wallet_name=party.wallet_name,
                    key_name=party.key_name,
                    market_name=self.market_name
                    if self.market_name is not None
                    else self.market_manager_args.market_name,
                    asset_name=self.market_manager_args.asset_name,
                    price_process_generator=iter(self.price_process),
                    initial_asset_mint=self.sensitive_trader_args.initial_mint,
                    scale=self.sensitive_trader_args.scale[i],
                    max_order_size=self.sensitive_trader_args.max_order_size[i],
                    tag=i,
                )
                for i, party in enumerate(SENSITIVE_TRADER_AGENTS)
            ]

            agents = (
                [market_manager, market_maker]
                + auction_pass_agents
                + random_market_order_traders
                + sensitive_limit_order_traders
            )

        else:
            agents = []

        if kwargs.get("agent", None) is not None:
            # If a market and asset were not specified, use scenario config values
            if kwargs["agent"].market_name is None:
                kwargs["agent"].market_name = self.market_manager_args.market_name
            if kwargs["agent"].asset_name is None:
                kwargs["agent"].asset_name = self.market_manager_args.asset_name
            # If a sim, overwrite price process
            if kwargs.get("network", Network.FAIRGROUND) == Network.NULLCHAIN:
                kwargs["agent"].price_process_generator = iter(self.price_process)

            kwargs["agent"].order_validity_length = self.step_length_seconds

            agents.append(kwargs.get("agent", None))

        return {agent.name(): agent for agent in agents}

    def configure_environment(
        self,
        vega: Union[VegaServiceNull, VegaServiceNetwork],
        tag: Optional[str] = None,
        random_state: Optional[np.random.RandomState] = None,
        **kwargs,
    ) -> Union[MarketEnvironmentWithState, NetworkEnvironment]:
        if kwargs.get("network", Network.FAIRGROUND) == Network.NULLCHAIN:
            env = MarketEnvironmentWithState(
                agents=list(self.agents.values()),
                n_steps=self.simulation_args.n_steps,
                vega_service=vega,
                step_length_seconds=self.step_length_seconds,
                block_length_seconds=1,
                random_state=random_state,
                transactions_per_block=100,
            )
        else:
            env = NetworkEnvironment(
                agents=list(self.agents.values()),
                n_steps=-1,
                vega_service=vega,
                step_length_seconds=self.step_length_seconds,
                raise_datanode_errors=kwargs.get("raise_datanode_errors", False),
                raise_step_errors=kwargs.get("raise_step_errors", False),
                random_state=random_state,
            )
        return env



from vega_sim.parameter_test.parameter.experiment import SingleParameterExperiment
from vega_sim.parameter_test.parameter.loggers import (
    ideal_market_maker_single_data_extraction,
    target_stake_additional_data,
    tau_scaling_additional_data,
    limit_order_book,
)
from vega_sim.scenario.parameter_experiment.scenario import ParameterExperiment
from vega_sim.parameter_test.parameter.loggers import (
    BASE_IDEAL_MM_CSV_HEADERS,
    LOB_CSV_HEADERS,
)
from vega_sim.parameter_test.parameter.experiment import (
    FILE_PATTERN,
    FILE_PATTERN_LOB,
)

TARGET_STAKE_SCALING_FACTOR = SingleParameterExperiment(
    name="TargetStakeScalingFactor",
    parameter_type="market",
    parameter_to_vary=(
        "liquidity_monitoring_parameters.target_stake_parameters.scaling_factor"
    ),
    values=[0.5, 5, 50],
    scenario=ParameterExperiment(
        state_extraction_fn=ideal_market_maker_single_data_extraction(
            additional_data_fns=[
                tau_scaling_additional_data,
                target_stake_additional_data,
                limit_order_book,
            ]
        ),
    ),
    runs_per_scenario=2,
    additional_market_parameters_to_set={
        "liquidity_monitoring_parameters.triggering_ratio": "1",
    },
    data_extraction=[
        (FILE_PATTERN, BASE_IDEAL_MM_CSV_HEADERS),
        (FILE_PATTERN_LOB, LOB_CSV_HEADERS),
    ],
)

TAU_SCALING_FACTOR = SingleParameterExperiment(
    name="TauScalingFactor",
    parameter_type="network",
    parameter_to_vary="market.liquidity.probabilityOfTrading.tau.scaling",
    values=["10", "100", "1000"],
    scenario=ParameterExperiment(
        state_extraction_fn=ideal_market_maker_single_data_extraction(
            additional_data_fns=[
                tau_scaling_additional_data,
                target_stake_additional_data,
                limit_order_book,
            ]
        ),
    ),
    runs_per_scenario=2,
    additional_market_parameters_to_set={
        "liquidity_monitoring_parameters.triggering_ratio": "1",
    },
    data_extraction=[
        (FILE_PATTERN, BASE_IDEAL_MM_CSV_HEADERS),
        (FILE_PATTERN_LOB, LOB_CSV_HEADERS),
    ],
)

BOND_PENALTY_PARAMETER = SingleParameterExperiment(
    name="BondPenaltyFactor",
    parameter_type="network",
    parameter_to_vary="market.liquidity.bondPenaltyParameter",
    values=["0.0", "0.5", "1.0"],
    scenario=ParameterExperiment(
        state_extraction_fn=ideal_market_maker_single_data_extraction(
            additional_data_fns=[
                tau_scaling_additional_data,
                target_stake_additional_data,
                limit_order_book,
            ]
        ),
    ),
    runs_per_scenario=2,
    additional_market_parameters_to_set={
        "liquidity_monitoring_parameters.triggering_ratio": "1",
    },
    data_extraction=[
        (FILE_PATTERN, BASE_IDEAL_MM_CSV_HEADERS),
        (FILE_PATTERN_LOB, LOB_CSV_HEADERS),
    ],
)

CONFIGS = [
    TARGET_STAKE_SCALING_FACTOR,
    TAU_SCALING_FACTOR,
    BOND_PENALTY_PARAMETER,
]



import argparse
import datetime
import logging
import numpy as np
from typing import Any, Callable, List, Optional, Dict
from vega_sim.environment.agent import Agent
from vega_sim.scenario.scenario import Scenario
from vega_sim.environment.environment import RealtimeMarketEnvironment
from vega_sim.null_service import VegaServiceNull
from vega_sim.scenario.ideal_market_maker_v2.agents import (
    MM_WALLET,
    TERMINATE_WALLET,
)
from vega_sim.scenario.common.agents import MarketManager, StateAgent, KeyFunder

logger = logging.getLogger(__name__)


class RealtimeMarket(Scenario):
    def __init__(
        self,
        market_decimal: int = 5,
        asset_decimal: int = 5,
        market_position_decimal: int = 2,
        market_names: Optional[list[str]] = None,
        asset_names: Optional[list[str]] = None,
        block_size: int = 1,
        block_length_seconds: int = 1,
        state_extraction_fn: Optional[
            Callable[[VegaServiceNull, Dict[str, Agent]], Any]
        ] = None,
        pubkeys_to_faucet: Optional[list[str]] = None,
        amount_to_faucet: float = 100_000,
    ):
        super().__init__(state_extraction_fn=state_extraction_fn)
        if len(market_names) != len(asset_names):
            raise Exception("Must provide equal numbers of markets and assets")
        self.market_decimal = market_decimal
        self.asset_decimal = asset_decimal
        self.market_position_decimal = market_position_decimal

        self.block_size = block_size
        self.block_length_seconds = block_length_seconds
        self.market_names = ["ETH:USD"] if market_names is None else market_names
        self.asset_names = ["tDAI"] if asset_names is None else asset_names
        self.pubkeys_to_faucet = (
            pubkeys_to_faucet if pubkeys_to_faucet is not None else []
        )
        self.amount_to_faucet = amount_to_faucet

    def configure_agents(
        self,
        vega: VegaServiceNull,
        tag: Optional[str],
        random_state: Optional[np.random.RandomState],
        **kwargs,
    ) -> List[StateAgent]:
        # Set up market name and settlement asset

        agents = []
        for i, market_name in enumerate(self.market_names):
            agents.append(
                MarketManager(
                    key_name=MM_WALLET.name,
                    terminate_key_name=TERMINATE_WALLET.name,
                    asset_decimal=self.asset_decimal,
                    market_decimal=self.market_decimal,
                    market_position_decimal=self.market_position_decimal,
                    market_name=market_name,
                    asset_name=self.asset_names[i],
                    tag=market_name,
                )
            )

        for asset in self.asset_names:
            agents.append(
                KeyFunder(
                    self.pubkeys_to_faucet,
                    asset_to_fund=asset,
                    amount_to_fund=self.amount_to_faucet,
                    tag=asset,
                )
            )

        return {agent.name(): agent for agent in agents}

    def configure_environment(
        self,
        vega: VegaServiceNull,
        **kwargs,
    ) -> RealtimeMarketEnvironment:
        return RealtimeMarketEnvironment(
            agents=list(self.agents.values()),
            random_agent_ordering=True,
            transactions_per_block=self.block_size,
            vega_service=vega,
            block_length_seconds=self.block_length_seconds,
            pause_every_n_steps=None,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    scenario = RealtimeMarket(
        market_names=["ETH", "ETH2", "ETH3", "ETH4", "ETH5"],
        asset_names=["USD", "USD", "USD", "USD", "USD"],
        pubkeys_to_faucet=[
            "520c6b790113062ba29b6e689694766e3a0dbcbcc4d79e3a43e3f9d21fc3ec7b",
            "6001c9b8ce0b8577e2d407ec1584ff6e2cccf84e36404d8274b9da1b86d31022",
        ],
        market_decimal=2,
        asset_decimal=4,
        market_position_decimal=4,
        amount_to_faucet=10_000_000,
    )

    with VegaServiceNull(
        warn_on_raw_data_access=False,
        run_with_console=True,
        use_full_vega_wallet=False,
        retain_log_files=True,
        launch_graphql=False,
        seconds_per_block=1,  # Heuristic
        transactions_per_block=2000,
    ) as vega:
        scenario.run_iteration(
            vega=vega,
            pause_at_completion=True,
        )



from typing import Dict, Type, Optional, Union
from dataclasses import dataclass
from vega_sim.reinforcement.v2.agents.puppets import (
    AGENT_TYPE_TO_AGENT,
    AgentType,
    MarketOrderAction,
    Action,
    Side,
    Puppet,
)
from vega_sim.scenario.scenario import Scenario
from vega_sim.scenario.common.agents import MarketManager
from vega_sim.reinforcement.v2.rewards import BaseRewarder, REWARD_ENUM_TO_CLASS, Reward
from vega_sim.reinforcement.v2.states import State, SimpleState
from vega_sim.null_service import VegaServiceNull


@dataclass
class StepResult:
    observation: State
    reward: float


class Environment:
    def __init__(
        self,
        agents: Dict[str, AgentType],
        agent_to_reward: Union[Reward, Dict[str, Reward]],
        agent_to_state: Dict[str, Type[State]],
        scenario: Scenario,
        reset_vega_every_n_runs: int = 100,
        funds_per_run: float = 10_000,
    ):
        self._agents = agents
        self._agent_to_state = agent_to_state

        self._agent_to_reward: Dict[str, BaseRewarder] = {}

        if isinstance(agent_to_reward, dict):
            self._is_single_reward = False
            self._agent_to_reward_enum = agent_to_reward
        else:
            self._is_single_reward = True
            self._single_reward_base: Reward = agent_to_reward

        self._scenario = scenario
        self._vega = VegaServiceNull(
            warn_on_raw_data_access=False,
            run_with_console=False,
            retain_log_files=True,
            store_transactions=True,
            transactions_per_block=1000,
        )
        self._loop_tag = 0

        self._runs_since_reset = 0
        self._reset_vega_every_n_runs = reset_vega_every_n_runs
        self._funds_per_run = funds_per_run

        self._vega.start()

    def stop(self):
        self._vega.stop()

    def _extract_observation(self, agent_name: str) -> State:
        return self._agent_to_state[agent_name].from_vega(
            self._vega,
            for_key_name=agent_name,
            market_id=self._market_id,
            asset_id=self._asset_id,
        )

    def step(self, actions: Dict[str, Optional[Action]]) -> Dict[str, StepResult]:
        for agent_name, action in actions.items():
            if action is not None:
                self._puppets[agent_name].set_next_action(action=action)

        self._scenario.env.step(self._vega)
        self._vega.wait_fn(1)
        step_res = {}

        for agent_name, reward_gen in self._agent_to_reward.items():
            step_res[agent_name] = StepResult(
                observation=self._extract_observation(agent_name),
                reward=self.calculate_reward(reward_gen),
            )
        return step_res

    def calculate_reward(self, rewarder: Type[BaseRewarder]) -> float:
        return rewarder.get_reward(vega=self._vega)

    def _reset_vega(self) -> None:
        self._vega.stop()
        self._vega = VegaServiceNull(
            warn_on_raw_data_access=False,
            run_with_console=False,
            retain_log_files=True,
            store_transactions=True,
            transactions_per_block=1000,
        )
        self._vega.start()

    def reset(self) -> None:
        if self._runs_since_reset > self._reset_vega_every_n_runs:
            self._reset_vega()
            self._runs_since_reset = 0
        self._runs_since_reset += 1

        self._scenario.agents = self._scenario.configure_agents(
            vega=self._vega, tag=str(self._loop_tag), random_state=None
        )
        manager = self._scenario.agents[
            MarketManager.name_from_tag(str(self._loop_tag))
        ]

        self._puppets: Dict[str, Puppet] = {
            agent_name: AGENT_TYPE_TO_AGENT[agent_type](
                key_name=agent_name,
                market_name=manager.market_name,
                tag=str(self._loop_tag),
            )
            for agent_name, agent_type in self._agents.items()
        }
        self._scenario.agents.update(self._puppets)

        self._scenario.env = self._scenario.configure_environment(
            vega=self._vega,
            tag=str(self._loop_tag),
        )

        for agent in self._scenario.env.agents:
            agent.initialise(vega=self._vega)
            self._vega.wait_fn(1)

        for agent_name in self._agents.keys():
            self._vega.mint(
                key_name=agent_name,
                asset=manager.asset_id,
                amount=self._funds_per_run,
            )

        self._agent_to_reward = {}

        if not self._is_single_reward:
            for agent, reward in self._agent_to_reward_enum.items():
                self._agent_to_reward[agent] = REWARD_ENUM_TO_CLASS[reward](
                    agent_key=agent,
                    asset_id=manager.asset_id,
                    market_id=manager.market_id,
                )
        else:
            for agent in self._agent_to_state.keys():
                self._agent_to_reward[agent] = REWARD_ENUM_TO_CLASS[
                    self._single_reward_base
                ](
                    agent_keys=list(self._agent_to_state.keys()),
                    asset_id=manager.asset_id,
                    market_id=manager.market_id,
                )
        self._market_id = manager.market_id
        self._asset_id = manager.asset_id


if __name__ == "__main__":
    from vega_sim.scenario.registry import CurveMarketMaker

    scenario = CurveMarketMaker(
        market_decimal=3,
        asset_decimal=5,
        market_position_decimal=2,
        initial_price=1000.0,
        lp_commitamount=100000,
        initial_asset_mint=1e8,
        step_length_seconds=1,
        block_length_seconds=1,
        buy_intensity=5,
        sell_intensity=5,
        market_name="ETH",
        num_steps=50,
        random_agent_ordering=False,
        sigma=100,
        asset_name="DAI",
    )

    env = Environment(
        agents={"test_agent_learner": AgentType.MARKET_ORDER},
        agent_to_reward={"test_agent_learner": Reward.PNL},
        agent_to_state={"test_agent_learner": SimpleState},
        scenario=scenario,
    )
    env.reset()
    for _ in range(10):
        print(env.step({"test_agent_learner": MarketOrderAction(Side.BUY, 1)}))

    env.stop()



from vega_sim.api.market import MarketConfig

import argparse
import logging
import numpy as np
from typing import Optional, List
from vega_sim.scenario.common.utils.price_process import random_walk
from collections import defaultdict

from vega_sim.scenario.scenario import Scenario
from vega_sim.environment.environment import (
    MarketEnvironmentWithState,
    NetworkEnvironment,
)
from vega_sim.scenario.constants import Network
from vega_sim.null_service import VegaServiceNull

from vega_sim.scenario.configurable_market.agents import ConfigurableMarketManager
from vega_sim.scenario.common.agents import (
    StateAgent,
    UncrossAuctionAgent,
    ExponentialShapedMarketMaker,
    MarketOrderTrader,
    LimitOrderTrader,
    RewardFunder,
    AtTheTouchMarketMaker,
)
from vega_sim.scenario.fuzzed_markets.agents import (
    FuzzySuccessorConfigurableMarketManager,
    FuzzingAgent,
    RiskyMarketOrderTrader,
    RiskySimpleLiquidityProvider,
    FuzzyLiquidityProvider,
)

import vega_sim.proto.vega as vega_protos

import datetime
from typing import Optional, Dict
from dataclasses import dataclass
from vega_sim.scenario.common.agents import ExponentialShapedMarketMaker
import pandas as pd


@dataclass
class MarketHistoryAdditionalData:
    at_time: datetime.datetime
    external_prices: Dict[str, float]
    trader_close_outs: Dict[str, int]
    liquidity_provider_close_outs: Dict[str, int]


def state_extraction_fn(vega: VegaServiceNull, agents: dict):
    at_time = vega.get_blockchain_time()

    external_prices = {}
    trader_close_outs = {}
    liquidity_provider_close_outs = {}

    for _, agent in agents.items():
        if isinstance(agent, ExponentialShapedMarketMaker):
            external_prices[agent.market_id] = agent.curr_price
        if isinstance(agent, RiskyMarketOrderTrader):
            trader_close_outs[agent.market_id] = (
                trader_close_outs.get(agent.market_id, 0) + agent.close_outs
            )
        if isinstance(agent, RiskySimpleLiquidityProvider):
            liquidity_provider_close_outs[agent.market_id] = (
                liquidity_provider_close_outs.get(agent.market_id, 0) + agent.close_outs
            )

    return MarketHistoryAdditionalData(
        at_time=at_time,
        external_prices=external_prices,
        trader_close_outs=trader_close_outs,
        liquidity_provider_close_outs=liquidity_provider_close_outs,
    )


def additional_data_to_rows(data) -> List[pd.Series]:
    results = []
    for market_id in data.external_prices.keys():
        results.append(
            {
                "time": data.at_time,
                "market_id": market_id,
                "external_price": data.external_prices.get(market_id, np.NaN),
                "trader_close_outs": data.trader_close_outs.get(market_id, 0),
                "liquidity_provider_close_outs": data.liquidity_provider_close_outs.get(
                    market_id, 0
                ),
            }
        )
    return results


def _create_price_process(
    random_state: np.random.RandomState, num_steps, decimal_places
):
    price_process = [1500]

    while len(price_process) < num_steps + 1:
        # Add a stable price-process with a random duration of 5-20% of the sim
        price_process = np.concatenate(
            (
                price_process,
                random_walk(
                    num_steps=random_state.randint(
                        int(0.05 * num_steps), int(0.20 * num_steps)
                    ),
                    sigma=2,
                    starting_price=price_process[-1],
                    decimal_precision=decimal_places,
                ),
            )
        )
        # Add an unstable price-process with a random duration of 5-10% of the sim
        price_process = np.concatenate(
            (
                price_process,
                random_walk(
                    num_steps=random_state.randint(
                        int(0.05 * num_steps), int(0.10 * num_steps)
                    ),
                    sigma=2,
                    drift=random_state.uniform(-1, 1),
                    starting_price=price_process[-1],
                    decimal_precision=decimal_places,
                ),
            )
        )

    # Add spikes to the price monitoring auction
    spike_at = random_state.randint(0, num_steps, size=3)
    for i in spike_at:
        price_process[i] = price_process[i] * random_state.choice([0.95, 1.05])

    return price_process


class FuzzingScenario(Scenario):
    def __init__(
        self,
        num_steps: int = 60 * 24 * 30 * 3,
        transactions_per_block: int = 4096,
        block_length_seconds: float = 1,
        n_markets: int = 2,
        step_length_seconds: Optional[float] = None,
        fuzz_market_config: Optional[dict] = None,
        output: bool = True,
    ):
        super().__init__(
            state_extraction_fn=lambda vega, agents: state_extraction_fn(vega, agents),
            additional_data_output_fns={
                "additional_data.csv": lambda data: additional_data_to_rows(data),
            },
        )

        self.n_markets = n_markets
        self.fuzz_market_config = fuzz_market_config

        self.num_steps = num_steps
        self.step_length_seconds = (
            step_length_seconds
            if step_length_seconds is not None
            else block_length_seconds
        )

        self.block_length_seconds = block_length_seconds
        self.transactions_per_block = transactions_per_block

        self.output = output

    def configure_agents(
        self,
        vega: VegaServiceNull,
        tag: str,
        random_state: Optional[np.random.RandomState],
        **kwargs,
    ) -> List[StateAgent]:
        self.random_state = (
            random_state if random_state is not None else np.random.RandomState()
        )

        self.all_agents = {
            "market_managers": [],
            "market_makers": [],
            "at_touch_market_makers": [],
            "auction_traders": [],
            "random_traders": [],
            "fuzz_traders": [],
            "risky_traders": [],
            "fuzz_liquidity_providers": [],
            "risky_liquidity_providers": [],
            "reward_funders": [],
        }

        self.market_agents = {}
        self.initial_asset_mint = 10e9

        for i_market in range(self.n_markets):
            # Define the market and the asset:
            market_name = f"ASSET_{str(i_market).zfill(3)}"
            asset_name = f"ASSET_{str(i_market).zfill(3)}"
            asset_dp = 18

            market_agents = {}
            # Create fuzzed market config
            market_config = MarketConfig()
            market_config.set(
                "liquidity_monitoring_parameters.target_stake_parameters.scaling_factor",
                1e-4,
            )
            if self.fuzz_market_config is not None:
                for param in self.fuzz_market_config:
                    market_config.set(param=self.fuzz_market_config[param])

            # Create fuzzed price process
            price_process = _create_price_process(
                random_state=self.random_state,
                num_steps=self.num_steps,
                decimal_places=int(market_config.decimal_places),
            )

            # Create fuzzed market managers
            market_agents["market_managers"] = [
                FuzzySuccessorConfigurableMarketManager(
                    proposal_wallet_name="MARKET_MANAGER",
                    proposal_key_name="PROPOSAL_KEY",
                    termination_wallet_name="MARKET_MANAGER",
                    termination_key_name="TERMINATION_KEY",
                    market_config=market_config,
                    market_name=market_name,
                    market_code=market_name,
                    asset_dp=asset_dp,
                    asset_name=asset_name,
                    settlement_price=price_process[-1],
                    stake_key=True if kwargs["network"] == Network.CAPSULE else False,
                    tag=f"MARKET_{str(i_market).zfill(3)}",
                    market_agents=market_agents,
                    successor_probability=0.01,
                )
            ]

            market_agents["market_makers"] = [
                ExponentialShapedMarketMaker(
                    wallet_name="MARKET_MAKERS",
                    key_name=f"MARKET_{str(i_market).zfill(3)}",
                    price_process_generator=iter(price_process),
                    initial_asset_mint=self.initial_asset_mint,
                    market_name=market_name,
                    asset_name=asset_name,
                    commitment_amount=1e6,
                    market_decimal_places=market_config.decimal_places,
                    asset_decimal_places=asset_dp,
                    num_steps=self.num_steps,
                    kappa=2.4,
                    tick_spacing=0.05,
                    market_kappa=50,
                    state_update_freq=10,
                    tag=f"MARKET_{str(i_market).zfill(3)}",
                )
            ]

            market_agents["at_touch_market_makers"] = [
                AtTheTouchMarketMaker(
                    wallet_name=f"AT_THE_TOUCH_MM",
                    key_name=f"MARKET_{str(i_market).zfill(3)}",
                    initial_asset_mint=1e5,
                    market_name=market_name,
                    asset_name=asset_name,
                    order_size=1,
                    peg_offset=0,
                    max_position=5,
                    tag=f"MARKET_{str(i_market).zfill(3)}",
                )
            ]

            market_agents["auction_traders"] = [
                UncrossAuctionAgent(
                    wallet_name="AUCTION_TRADERS",
                    key_name=f"MARKET_{str(i_market).zfill(3)}_{side}",
                    side=side,
                    initial_asset_mint=self.initial_asset_mint,
                    price_process=iter(price_process),
                    market_name=market_name,
                    asset_name=asset_name,
                    uncrossing_size=20,
                    tag=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                )
                for i_agent, side in enumerate(["SIDE_BUY", "SIDE_SELL"])
            ]

            market_agents["random_traders"] = [
                MarketOrderTrader(
                    wallet_name="RANDOM_TRADERS",
                    key_name=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                    market_name=market_name,
                    asset_name=asset_name,
                    buy_intensity=10,
                    sell_intensity=10,
                    base_order_size=1,
                    step_bias=1,
                    tag=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                )
                for i_agent in range(5)
            ]

            for i_agent in range(5):
                market_agents["random_traders"].append(
                    LimitOrderTrader(
                        wallet_name=f"RANDOM_TRADERS",
                        key_name=f"LIMIT_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                        market_name=market_name,
                        asset_name=asset_name,
                        time_in_force_opts={"TIME_IN_FORCE_GTT": 1},
                        buy_volume=1,
                        sell_volume=1,
                        buy_intensity=10,
                        sell_intensity=10,
                        submit_bias=1,
                        cancel_bias=0,
                        duration=120,
                        price_process=price_process,
                        spread=0,
                        mean=-3,
                        sigma=0.5,
                        tag=f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                    )
                )

            market_agents["fuzz_traders"] = [
                FuzzingAgent(
                    wallet_name="FUZZING_TRADERS",
                    key_name=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                    market_name=market_name,
                    asset_name=asset_name,
                    output_plot_on_finalise=self.output,
                    tag=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                )
                for i_agent in range(10)
            ]

            market_agents["risky_traders"] = [
                RiskyMarketOrderTrader(
                    wallet_name="risky_traders",
                    key_name=f"MARKET_{str(i_market).zfill(3)}_SIDE_{side}_AGENT_{str(i_agent).zfill(3)}",
                    market_name=market_name,
                    asset_name=asset_name,
                    side=side,
                    initial_asset_mint=1_000,
                    size_factor=0.5,
                    step_bias=0.1,
                    tag=f"MARKET_{str(i_market).zfill(3)}_SIDE_{side}_AGENT_{str(i_agent).zfill(3)}",
                )
                for side in ["SIDE_BUY", "SIDE_SELL"]
                for i_agent in range(10)
            ]

            market_agents["risky_liquidity_providers"] = [
                RiskySimpleLiquidityProvider(
                    wallet_name="risky_liquidity_providers",
                    key_name=f"HIGH_RISK_LPS_MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                    market_name=market_name,
                    asset_name=asset_name,
                    initial_asset_mint=20_000,
                    commitment_factor=0.5,
                    step_bias=0.1,
                    tag=f"HIGH_RISK_LPS_MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                )
                for i_agent in range(5)
            ]

            for i_agent in range(45):
                market_agents["risky_liquidity_providers"].append(
                    RiskySimpleLiquidityProvider(
                        wallet_name="risky_liquidity_providers",
                        key_name=f"LOW_RISK_LPS_MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                        market_name=market_name,
                        asset_name=asset_name,
                        initial_asset_mint=20_000,
                        commitment_factor=0.3,
                        step_bias=0.1,
                        tag=f"LOW_RISK_LPS_MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                    )
                )

            market_agents["fuzz_liquidity_providers"] = [
                FuzzyLiquidityProvider(
                    wallet_name="FUZZY_LIQUIDITY_PROVIDERS",
                    key_name=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                    market_name=market_name,
                    asset_name=asset_name,
                    initial_asset_mint=5_000,
                    tag=(
                        f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}"
                    ),
                )
            ]

            market_agents["reward_funders"] = []
            for i_agent, (account_type, metric) in enumerate(
                [
                    (
                        vega_protos.vega.ACCOUNT_TYPE_REWARD_MAKER_PAID_FEES,
                        vega_protos.vega.DISPATCH_METRIC_MAKER_FEES_PAID,
                    ),
                    (
                        vega_protos.vega.ACCOUNT_TYPE_REWARD_MAKER_RECEIVED_FEES,
                        vega_protos.vega.DISPATCH_METRIC_MAKER_FEES_RECEIVED,
                    ),
                    (
                        vega_protos.vega.ACCOUNT_TYPE_REWARD_LP_RECEIVED_FEES,
                        vega_protos.vega.DISPATCH_METRIC_LP_FEES_RECEIVED,
                    ),
                    (
                        vega_protos.vega.ACCOUNT_TYPE_REWARD_MARKET_PROPOSERS,
                        vega_protos.vega.DISPATCH_METRIC_MARKET_VALUE,
                    ),
                ]
            ):
                market_agents["reward_funders"].append(
                    RewardFunder(
                        wallet_name="REWARD_FUNDERS",
                        key_name=f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                        reward_asset_name=vega_protos.vega.DispatchMetric.Name(metric),
                        initial_mint=1e9,
                        account_type=account_type,
                        transfer_amount=100,
                        asset_for_metric_name=asset_name,
                        metric=metric,
                        market_names=[market_name],
                        stake_key=True,
                        tag=f"MARKET_{str(i_market).zfill(3)}_AGENT_{str(i_agent).zfill(3)}",
                    )
                )
            self.market_agents[market_name] = market_agents
            for agent_type, agent_list in market_agents.items():
                self.all_agents[agent_type].extend(agent_list)

        return {
            agent.name(): agent
            for agent_list in self.all_agents.values()
            for agent in agent_list
        }

    def configure_environment(
        self,
        vega: VegaServiceNull,
        **kwargs,
    ) -> MarketEnvironmentWithState:
        if kwargs.get("network", Network.NULLCHAIN) == Network.NULLCHAIN:
            return MarketEnvironmentWithState(
                agents=list(self.agents.values()),
                n_steps=self.num_steps,
                random_agent_ordering=False,
                transactions_per_block=self.transactions_per_block,
                vega_service=vega,
                step_length_seconds=self.step_length_seconds,
                block_length_seconds=vega.seconds_per_block,
            )
        else:
            return NetworkEnvironment(
                agents=list(self.agents.values()),
                n_steps=self.num_steps,
                vega_service=vega,
                step_length_seconds=self.step_length_seconds,
                raise_datanode_errors=kwargs.get("raise_datanode_errors", False),
                raise_step_errors=kwargs.get("raise_step_errors", False),
                random_state=self.random_state,
                create_keys=True,
                mint_keys=True,
            )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    scenario = FuzzingScenario(
        num_steps=10000,
        step_length_seconds=5,
        block_length_seconds=1,
        transactions_per_block=4096,
    )

    with VegaServiceNull(
        warn_on_raw_data_access=False,
        run_with_console=True,
        use_full_vega_wallet=False,
        retain_log_files=True,
        launch_graphql=True,
        seconds_per_block=scenario.block_length_seconds,
        transactions_per_block=scenario.transactions_per_block,
    ) as vega:
        scenario.run_iteration(
            vega=vega,
            pause_at_completion=False,
        )



import logging

import gymnasium as gym
from gymnasium import spaces
from typing import Type, Optional
from enum import Enum
from stable_baselines3.common.callbacks import BaseCallback

from vega_sim.reinforcement.v2.agents.puppets import (
    AgentType,
    ForcedSide,
    Side,
    AGENT_TYPE_TO_ACTION,
)
from vega_sim.reinforcement.v2.learning_environment import Environment
from vega_sim.reinforcement.v2.rewards import Reward, REWARD_ENUM_TO_CLASS
from vega_sim.reinforcement.v2.stable_baselines.states import (
    price_state_with_fees_obs_space,
    position_state_with_fees_obs_space,
)
from vega_sim.reinforcement.v2.states import PriceStateWithFees, State, PositionOnly
from vega_sim.scenario.registry import CurveMarketMaker


logger = logging.getLogger(__name__)


class ActionType(Enum):
    MARKET = "market"
    AT_TOUCH_ONE_SIDE = "at_touch_one_side"


ACTION_TO_AGENT = {
    ActionType.MARKET: AgentType.MARKET_ORDER,
    ActionType.AT_TOUCH_ONE_SIDE: AgentType.AT_TOUCH,
}


class ActionLoggerCallback(BaseCallback):
    pass


class SingleAgentVegaEnv(gym.Env):
    """Custom Environment that follows gym interface."""

    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        action_type: ActionType = ActionType.MARKET,
        state_type: Type[State] = PriceStateWithFees,
        reward_type: Reward = Reward.PNL,
        num_levels_state: int = 5,
        trade_volume: float = 1,
        steps_per_trading_session: int = 1000,
        terminal_reward_type: Optional[Reward] = None,
    ):
        super().__init__()
        self.num_levels_state = num_levels_state
        self.state_type = state_type
        self.action_type = action_type
        self.trade_volume = trade_volume
        self.steps_per_trading_session = steps_per_trading_session
        self.current_step = 0
        self.learner_name = "learner_1"
        self.terminal_reward_type = terminal_reward_type

        # Define action and observation space
        # They must be gym.spaces objects
        # Example when using discrete actions:
        self.action_space = self._get_action_space(action_type=action_type)
        # Example for using image as input (channel-first; channel-last also works):
        self.observation_space = self._get_observation_space(self.state_type)

        scenario = CurveMarketMaker(
            market_decimal=3,
            asset_decimal=5,
            market_position_decimal=2,
            initial_price=1000.0,
            lp_commitamount=100000,
            initial_asset_mint=1e8,
            step_length_seconds=1,
            block_length_seconds=1,
            # market_maker_assumed_market_kappa=0.8,
            buy_intensity=5,
            sell_intensity=5,
            market_name="ETH",
            num_steps=self.steps_per_trading_session * 2,
            random_agent_ordering=False,
            sigma=100,
            asset_name="DAI",
        )

        self.env = Environment(
            agents={self.learner_name: ACTION_TO_AGENT[self.action_type]},
            agent_to_reward={self.learner_name: reward_type},
            agent_to_state={self.learner_name: state_type},
            scenario=scenario,
        )

    def _get_action_space(self, action_type: ActionType) -> spaces.Space:
        if action_type == ActionType.MARKET:
            return spaces.Discrete(3)
        elif action_type == ActionType.AT_TOUCH_ONE_SIDE:
            return spaces.Discrete(2)
        else:
            raise Exception(f"Action type {action_type} is not implemented")

    def _get_observation_space(self, state_type: Type[State]) -> spaces.Space:
        if state_type == PriceStateWithFees:
            return price_state_with_fees_obs_space(num_levels=self.num_levels_state)
        if state_type == PositionOnly:
            return position_state_with_fees_obs_space()

    def _action_conversion(self, action):
        if self.action_type == ActionType.MARKET:
            return AGENT_TYPE_TO_ACTION[AgentType.MARKET_ORDER](
                side=Side(action), volume=self.trade_volume
            )
        elif self.action_type == ActionType.AT_TOUCH_ONE_SIDE:
            return AGENT_TYPE_TO_ACTION[AgentType.AT_TOUCH](
                side=ForcedSide(action), volume=self.trade_volume
            )
        else:
            raise Exception(f"Action type {self.action_type} is not implemented")

    def step(self, action):
        step_res = self.env.step(
            {self.learner_name: self._action_conversion(action=action)}
        )[self.learner_name]
        self.current_step += 1

        is_terminal = self.current_step >= self.steps_per_trading_session

        if is_terminal and self.terminal_reward_type is not None:
            terminal_reward = REWARD_ENUM_TO_CLASS[self.terminal_reward_type](
                agent_key=self.learner_name,
                asset_id=self.env._asset_id,
                market_id=self.env._market_id,
            )
            reward = step_res.reward + self.env.calculate_reward(terminal_reward)
        else:
            reward = step_res.reward

        return (
            step_res.observation.to_array(),
            reward,
            is_terminal,
            False,
            {},
        )

    def reset(self, seed=None):
        self.env.reset()
        step_res = self.env.step({self.learner_name: None})[self.learner_name]

        self.current_step = 0
        return step_res.observation.to_array(), {}

    def render(self, mode="human"):
        pass

    def close(self):
        self.env.stop()



import os
import logging
import json

from vega_sim.reinforcement.v2.agents.learning_agent import LearningAgent
from vega_sim.reinforcement.v2.agents.learning_agent_MO import LearningAgentFixedVol
from vega_sim.reinforcement.v2.agents.puppets import AgentType, NoAction
from vega_sim.reinforcement.v2.learning_environment import Environment
from vega_sim.reinforcement.v2.rewards import Reward
from vega_sim.reinforcement.v2.states import State, PriceStateWithFees
from vega_sim.scenario.registry import CurveMarketMaker

logger = logging.getLogger(__name__)


def run_training(
    agent_map: dict[str, LearningAgent],
    agent_type_map: dict[str, AgentType],
    agent_to_reward: dict[str, Reward],
    agent_to_state: dict[str, State],
    steps_per_iteration: int = 100,
    num_training_iterations: int = 1000,
):
    scenario = CurveMarketMaker(
        market_decimal=3,
        asset_decimal=5,
        market_position_decimal=2,
        initial_price=1000.0,
        lp_commitamount=100000,
        initial_asset_mint=1e8,
        step_length_seconds=1,
        block_length_seconds=1,
        buy_intensity=5,
        sell_intensity=5,
        market_name="ETH",
        num_steps=500,
        random_agent_ordering=False,
        sigma=100,
        asset_name="DAI",
    )

    env = Environment(
        agents=agent_type_map,
        agent_to_reward=agent_to_reward,
        agent_to_state=agent_to_state,
        scenario=scenario,
    )
    env.reset()
    result_dict = env.step({name: NoAction for name in agent_map.keys()})
    pas = []
    for i in range(num_training_iterations):
        for _ in range(steps_per_iteration):
            actions = {
                name: agent_map[name].step(result.observation)
                for (name, result) in result_dict.items()
            }
            result_dict = env.step(actions)
            for agent_name, agent in agent_map.items():
                result_val = result_dict[agent_name]
                agent.update_memory(
                    state=result_val.observation,
                    action=actions[agent_name],
                    reward=result_val.reward,
                )
        for agent in agent_map.values():
            agent.learning_step()

        logger.info(f"Completed training iteration {i}")

        pa = env._vega.party_account(
            key_name="test_agent_learner",
            asset_id=env._asset_id,
            market_id=env._market_id,
        )
        pa2 = env._vega.party_account(
            key_name="test_agent_learner_2",
            asset_id=env._asset_id,
            market_id=env._market_id,
        )
        pas.append(sum(pa))
        with open("bals.json", "w") as f:
            json.dump(pas, f)

        logger.info(f"Learner ended with {sum(pa)}")
        logger.info(f"Learner 2 ended with {sum(pa2)}")

        env.reset()

    env.stop()


def main():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s: %(message)s",
    )
    logfile_pol_imp = os.path.join("learning_pol_imp.csv")
    logfile_pol_eval = os.path.join("learning_pol_eval.csv")
    logfile_pnl = os.path.join("learning_pnl.csv")

    # set device
    device = "cpu"

    # create the Learning Agents
    agent_map = {
        "test_agent_learner": LearningAgentFixedVol(
            device=device,
            logfile_pol_imp=logfile_pol_imp,
            logfile_pol_eval=logfile_pol_eval,
            logfile_pnl=logfile_pnl,
            discount_factor=0.8,
            num_levels=5,
            inventory_penalty=0.1,
        ),
        "test_agent_learner_2": LearningAgentFixedVol(
            device=device,
            logfile_pol_imp=logfile_pol_imp + "_2",
            logfile_pol_eval=logfile_pol_eval + "_2",
            logfile_pnl=logfile_pnl,
            discount_factor=0.8,
            num_levels=5,
            inventory_penalty=0.1,
        ),
    }

    agents = {}
    agent_to_reward = {}
    agent_to_state = {}

    for tal in agent_map.keys():
        agents[tal] = AgentType.MARKET_ORDER
        agent_to_reward[tal] = Reward.PNL
        agent_to_state[tal] = PriceStateWithFees

    run_training(
        agent_map=agent_map,
        agent_type_map=agents,
        agent_to_reward=agent_to_reward,
        agent_to_state=agent_to_state,
    )


if __name__ == "__main__":
    main()
